import pytest
from demoproject.demoapp.models import AbstractSender, DemoModel, Sender1, Sender2
from strategy_field.registry import Registry
from strategy_field.utils import fqn


def test_registry():
    r = Registry(AbstractSender)
    r.register(Sender1)

    assert Sender1 in r
    assert fqn(Sender1) in r
    assert Sender2 not in r


def test_registry_check_classes():
    r = Registry(AbstractSender)
    with pytest.raises(ValueError):
        r.register(DemoModel)


def test_registry_bypass_class_check():
    r = Registry(None)
    r.register(DemoModel)
    r.register(AbstractSender)
    r.register(Sender1)

    assert Sender1 in r
    assert AbstractSender in r
    assert DemoModel in r


def test_registry_string():
    r = Registry("demoproject.demoapp.models.AbstractSender")
    r.register(Sender1)

    assert Sender1 in r
    assert fqn(Sender1) in r
    assert Sender2 not in r


def test_registry_is_valid():
    r = Registry("demoproject.demoapp.models.AbstractSender")

    assert r.is_valid(Sender1)
    assert r.is_valid(fqn(Sender1))
    assert not r.is_valid(DemoModel)

    r = Registry(None)
    assert r.is_valid(Sender1)
    assert r.is_valid(DemoModel)
    assert not r.is_valid("demoproject.demoapp.models.Wrong")


def test_registry_append():
    r = Registry("demoproject.demoapp.models.AbstractSender")

    assert r.register(Sender1)
    assert r.register(fqn(Sender2))
    assert not r.register("demoproject.demoapp.models.AbstractSender")


def test_registry_as_choices(monkeypatch):
    r = Registry("demoproject.demoapp.models.AbstractSender", label_attribute="label")

    r.register(Sender1)
    r.register(Sender2)
    r.register(fqn(Sender1))
    monkeypatch.setattr(Sender1, "label", classmethod(lambda s: "LABEL"), raising=False)

    assert r.as_choices() == [
        ("demoproject.demoapp.models.Sender1", "LABEL"),
        ("demoproject.demoapp.models.Sender2", "demoproject.demoapp.models.Sender2"),
    ]


def test_registry_contains():
    r = Registry("demoproject.demoapp.models.AbstractSender")

    r.register(Sender1)
    r.register(fqn(Sender2))
    assert Sender1 in r
    assert fqn(Sender1) in r
    assert "a.b.c" not in r



from django.contrib.auth.backends import ModelBackend
from django.contrib.auth.models import Permission


class AnyUserBackend(ModelBackend):
    supports_object_permissions = False
    supports_anonymous_user = True

    def get_all_permissions(self, user_obj, obj=None):
        # if settings.DEBUG:
        return (
            Permission.objects.all()
            .values_list("content_type__app_label", "codename")
            .order_by()
        )
        # return super().get_all_permissions(user_obj, obj)

    def get_group_permissions(self, user_obj, obj=None):
        # if settings.DEBUG:
        return (
            Permission.objects.all()
            .values_list("content_type__app_label", "codename")
            .order_by()
        )
        # return super().get_group_permissions(user_obj,
        # obj)

    def has_perm(self, user_obj, perm, obj=None):
        # if settings.DEBUG:
        return True
        # return super().has_perm(user_obj, perm, obj)

    def has_module_perms(self, user_obj, app_label):
        # if settings.DEBUG:
        return True
        # return super().has_module_perms(user_obj,
        # app_label)



# flake8: noqa
# noqa
import pytest
from demoproject.demoapp.models import DemoMultipleCustomModel, Strategy, Strategy1
from django.forms.models import modelform_factory
from django.urls import reverse
from strategy_field.utils import fqn


def pytest_generate_tests(metafunc):
    func_name = metafunc.function.__name__
    values = ids = []
    if "target" in metafunc.fixturenames:
        if func_name.endswith("_lookup_in"):
            values = [
                lambda o: [fqn(o.sender[0])],
                lambda o: o.sender,
                lambda o: [fqn(Strategy), fqn(Strategy1)],
                lambda o: [Strategy, Strategy1],
            ]
            ids = ["fqn(target.sender)", "target.sender", "fqn(Strategy)", "Strategy"]
        else:
            values = [lambda o: [fqn(Strategy)], lambda o: [Strategy]]
            ids = ["fqn(Strategy)", "Strategy"]

            if "demo_multiplecustom_model" in metafunc.fixturenames:
                values.extend([lambda o: [fqn(o.sender[0])], lambda o: o.sender])
                ids.extend(["fqn(target.sender)", "target.sender"])

        metafunc.parametrize("target", values, ids=ids)


def test_field():
    d = DemoMultipleCustomModel(sender=[Strategy])
    assert isinstance(d.sender[0], Strategy)


@pytest.mark.django_db
def test_model_save(target):
    d = DemoMultipleCustomModel(sender=target(None))
    d.save()
    assert isinstance(d.sender[0], Strategy)


@pytest.mark.django_db
def test_model_save_multiple():
    d = DemoMultipleCustomModel(sender=[Strategy, Strategy1])
    d.save()
    assert len(d.sender) == 2
    assert isinstance(d.sender[0], Strategy)
    assert isinstance(d.sender[1], Strategy1)


@pytest.mark.django_db
def test_model_get_or_create(target):
    d, __ = DemoMultipleCustomModel.objects.get_or_create(sender=target(None))
    assert isinstance(d.sender[0], Strategy)


@pytest.mark.django_db
def test_model_load(demo_multiplecustom_model):
    d = DemoMultipleCustomModel.objects.get(pk=demo_multiplecustom_model.pk)
    assert isinstance(d.sender[0], Strategy)


@pytest.mark.django_db
def test_form_load():
    d = DemoMultipleCustomModel(sender=[Strategy, Strategy1])
    form_class = modelform_factory(DemoMultipleCustomModel, exclude=[])
    form = form_class(instance=d)
    assert form.fields["sender"].choices == [
        ("demoproject.demoapp.models.Strategy", "demoproject.demoapp.models.Strategy"),
        (
            "demoproject.demoapp.models.Strategy1",
            "demoproject.demoapp.models.Strategy1",
        ),
    ]


@pytest.mark.django_db
def test_form_save(demo_multiplecustom_model):
    form_class = modelform_factory(DemoMultipleCustomModel, exclude=[])
    form = form_class(
        {"sender": [fqn(demo_multiplecustom_model.sender[0])]},
        instance=demo_multiplecustom_model,
    )
    form.is_valid()
    instance = form.save()
    assert fqn(instance.sender[0]) == fqn(demo_multiplecustom_model.sender[0])


@pytest.mark.django_db
def test_form_not_valid(demo_multiplecustom_model):
    form_class = modelform_factory(DemoMultipleCustomModel, exclude=[])
    form = form_class(
        {"sender": [fqn(DemoMultipleCustomModel)]}, instance=demo_multiplecustom_model
    )
    assert not form.is_valid()
    assert form.errors["sender"] == [
        "Select a valid choice. "
        "demoproject.demoapp.models.DemoMultipleCustomModel "
        "is not one of the available choices."
    ]


@pytest.mark.django_db
def test_form_default(demo_multiplecustom_model):
    form_class = modelform_factory(DemoMultipleCustomModel, exclude=[])
    form = form_class(instance=demo_multiplecustom_model)
    assert form.fields["sender"].choices == [
        ("demoproject.demoapp.models.Strategy", "demoproject.demoapp.models.Strategy"),
        (
            "demoproject.demoapp.models.Strategy1",
            "demoproject.demoapp.models.Strategy1",
        ),
    ]
    # assert form.as_table() == u'<tr><th><label for="id_sender">Sender:</label></th>' \
    #                           u'<td><select multiple="multiple" id="id_sender" name="sender">\n' \
    #                           u'<option value="demoproject.demoapp.models.Strategy" selected="selected">demoproject.demoapp.models.Strategy</option>\n' \
    #                           u'<option value="demoproject.demoapp.models.Strategy1">demoproject.demoapp.models.Strategy1</option>\n</select></td></tr>'


@pytest.mark.django_db
def test_admin_demo_multiple_model_add(webapp, admin_user):
    res = webapp.get("/demoapp/demomultiplecustommodel/add/", user=admin_user)
    form = res.forms["demomultiplecustommodel_form"]
    form["sender"].force_value(["demoproject.demoapp.models.Strategy"])
    form.submit().follow()
    assert (
        DemoMultipleCustomModel.objects.filter(
            sender="demoproject.demoapp.models.Strategy"
        ).count()
        == 1
    )


@pytest.mark.django_db
def test_admin_demo_multiple_model_edit(webapp, admin_user, demo_multiplecustom_model):
    demo_multiplecustom_model.sender = [Strategy, Strategy1]
    demo_multiplecustom_model.save()
    url = reverse(
        "admin:demoapp_demomultiplecustommodel_change",
        args=[demo_multiplecustom_model.pk],
    )
    res = webapp.get(url, user=admin_user)
    assert res.context["adminform"].form.fields["sender"].choices == [
        ("demoproject.demoapp.models.Strategy", "demoproject.demoapp.models.Strategy"),
        (
            "demoproject.demoapp.models.Strategy1",
            "demoproject.demoapp.models.Strategy1",
        ),
    ]

    form = res.forms["demomultiplecustommodel_form"]
    form["sender"] = [
        "demoproject.demoapp.models.Strategy",
        "demoproject.demoapp.models.Strategy1",
    ]
    form.submit().follow()
    res = webapp.get(url, user=admin_user)
    assert res.context["adminform"].form.fields["sender"].choices == [
        ("demoproject.demoapp.models.Strategy", "demoproject.demoapp.models.Strategy"),
        (
            "demoproject.demoapp.models.Strategy1",
            "demoproject.demoapp.models.Strategy1",
        ),
    ]

    # assert res.context['adminform'].form.as_table() == u'<tr><th><label for="id_sender">Sender:</label></th>' \
    #                                                    u'<td><select multiple="multiple" id="id_sender" name="sender">\n' \
    #                                                    u'<option value="demoproject.demoapp.models.Strategy" selected="selected">demoproject.demoapp.models.Strategy</option>\n' \
    #                                                    u'<option value="demoproject.demoapp.models.Strategy1" selected="selected">demoproject.demoapp.models.Strategy1</option>\n' \
    #                                                    u'</select></td></tr>'


# @pytest.mark.django_db
# def test_admin_demo_multiple_model_validate(webapp, admin_user, demo_multiplecustom_model):
# res = webapp.get('/admin/demo/demomultiplecustommodel/%s/' % demo_multiplecustom_model.pk, user=admin_user)
#     res.form['sender'] = ['invalid']
#     res = res.form.submit()
#     assert 'Select a valid choice' in res.context['adminform'].form.errors['sender'][0]


@pytest.mark.django_db
def test_demo_multiple_model_lookup_equal(demo_multiplecustom_model, target):
    assert (
        DemoMultipleCustomModel.objects.get(sender=target(demo_multiplecustom_model))
        == demo_multiplecustom_model
    )


@pytest.mark.django_db
def test_demo_multiple_model_lookup_contains(demo_multiplecustom_model, target):
    assert (
        DemoMultipleCustomModel.objects.get(
            sender__contains=target(demo_multiplecustom_model)
        )
        == demo_multiplecustom_model
    )


@pytest.mark.django_db
def test_demo_multiple_model_lookup_in(demo_multiplecustom_model, target):
    with pytest.raises(TypeError):
        assert (
            DemoMultipleCustomModel.objects.get(
                sender__in=target(demo_multiplecustom_model)
            )
            == demo_multiplecustom_model
        )



import pytest


@pytest.fixture
def registry():
    from demoproject.demoapp.models import AbstractSender, Sender1, Sender2
    from strategy_field.registry import Registry

    r = Registry(AbstractSender, label_attribute="label")
    r.register(Sender1)
    r.register(Sender2)
    return r


@pytest.fixture
def custom_registry():
    from demoproject.demoapp.models import Strategy
    from strategy_field.registry import Registry

    r = Registry(Strategy)
    r.register(Strategy)
    return r


@pytest.fixture
def demomodel():
    from demoproject.demoapp.models import DemoModel, Sender1

    return DemoModel.objects.get_or_create(sender=Sender1)[0]


@pytest.fixture
def democustommodel():
    from demoproject.demoapp.models import DemoCustomModel, Strategy
    from strategy_field.utils import fqn

    return DemoCustomModel.objects.get_or_create(sender=fqn(Strategy))[0]


@pytest.fixture
def demo_multiplecustom_model():
    from demoproject.demoapp.models import DemoMultipleCustomModel, Strategy
    from strategy_field.utils import fqn

    return DemoMultipleCustomModel.objects.get_or_create(sender=[fqn(Strategy)])[0]


@pytest.fixture
def demo_multiple_model():
    from demoproject.demoapp.models import DemoMultipleModel, Sender1

    return DemoMultipleModel.objects.get_or_create(sender=[Sender1])[0]


@pytest.fixture(scope="function")
def webapp(django_app):
    return django_app
    # import django_webtest
    #
    # wtm = django_webtest.WebTestMixin()
    # wtm.csrf_checks = False
    # wtm._patch_settings()
    # request.addfinalizer(wtm._unpatch_settings)


#     return django_webtest.DjangoTestApp()



# !qa: E501
import pytest
from demoproject.demoapp.models import DemoCustomModel, Strategy, Strategy1
from django.forms.models import modelform_factory
from django.urls import reverse
from strategy_field.utils import fqn


def pytest_generate_tests(metafunc):
    func_name = metafunc.function.__name__
    values = ids = []
    if "target_factory" in metafunc.fixturenames:
        if func_name.endswith("_lookup_in"):
            values = [
                lambda o: [fqn(o.sender)],
                lambda o: [o.sender],
                lambda o: [fqn(Strategy), fqn(Strategy1)],
                lambda o: [Strategy, Strategy1],
            ]
            ids = [
                fqn(Strategy),
                "<context.sender instance>",
                str([fqn(Strategy)]),
                str([Strategy, Strategy1]),
            ]
        elif func_name == ("test_model_save"):
            values = [
                lambda o: fqn(o.sender),
                lambda o: o.sender,
                lambda o: Strategy(o, "b"),
            ]
            ids = [fqn(Strategy), "<context.sender instance>", "Strategy1(context)"]
        else:
            values = [lambda o: fqn(Strategy)]
            ids = ["fqn(Strategy)"]
            if "democustommodel" in metafunc.fixturenames:
                values.extend([lambda o: fqn(o.sender), lambda o: o.sender])
                ids.extend(["fqn(target_factory.sender)", "target_factory.sender"])

        metafunc.parametrize("target_factory", values, ids=ids)


def test_field():
    d = DemoCustomModel(sender=Strategy)
    assert isinstance(d.sender, Strategy)
    assert d.sender == d.sender
    assert d.sender.context == d


@pytest.mark.django_db
def test_model_save(target_factory):
    d = DemoCustomModel(sender=Strategy)
    d.sender = target_factory(d)
    d.save()
    assert isinstance(d.sender, Strategy)
    assert d.sender == d.sender
    assert d.sender.context == d


@pytest.mark.django_db
def test_model_get_or_create(target_factory):
    d, __ = DemoCustomModel.objects.get_or_create(sender=target_factory(None))
    assert isinstance(d.sender, Strategy)


@pytest.mark.django_db
def test_model_load(democustommodel):
    d = DemoCustomModel.objects.get(pk=democustommodel.pk)
    assert isinstance(d.sender, Strategy)


# @pytest.mark.django_db
# def test_form(democustommodel, registry):
#     democustommodel._meta.get_field_by_name('sender')[0].registry = registry
#     form_class = modelform_factory(DemoCustomModel)
#     form = form_class(instance=democustommodel)
#     assert form.fields['sender'].choices[1:] == registry.as_choices()


@pytest.mark.django_db
def test_form_save(democustommodel):
    form_class = modelform_factory(DemoCustomModel, exclude=[])
    form = form_class({"sender": fqn(democustommodel.sender)}, instance=democustommodel)
    form.is_valid()
    instance = form.save()
    assert instance.sender == democustommodel.sender


@pytest.mark.django_db
def test_form_not_valid(democustommodel):
    form_class = modelform_factory(DemoCustomModel, exclude=[])
    form = form_class({"sender": fqn(DemoCustomModel)}, instance=democustommodel)
    assert not form.is_valid()
    assert form.errors["sender"] == [
        "Select a valid choice. "
        "demoproject.demoapp.models.DemoCustomModel "
        "is not one of the available choices."
    ]


@pytest.mark.django_db
def test_form_default(democustommodel):
    form_class = modelform_factory(DemoCustomModel, exclude=[])
    form = form_class(instance=democustommodel)
    assert form.fields["sender"].choices == [
        ("", "---------"),
        ("demoproject.demoapp.models.Strategy", "demoproject.demoapp.models.Strategy"),
        (
            "demoproject.demoapp.models.Strategy1",
            "demoproject.demoapp.models.Strategy1",
        ),
    ]


@pytest.mark.django_db
def test_admin_demomodel_add(webapp, admin_user):
    res = webapp.get("/demoapp/democustommodel/add/", user=admin_user)
    form = res.forms["democustommodel_form"]

    form["sender"] = "demoproject.demoapp.models.Strategy"
    form.submit().follow()
    assert (
        DemoCustomModel.objects.filter(
            sender="demoproject.demoapp.models.Strategy"
        ).count()
        == 1
    )


@pytest.mark.django_db
def test_admin_demomodel_edit(webapp, admin_user, democustommodel):
    url = reverse("admin:demoapp_democustommodel_change", args=[democustommodel.pk])
    res = webapp.get(url, user=admin_user)
    form = res.forms["democustommodel_form"]

    form["sender"] = "demoproject.demoapp.models.Strategy"
    form.submit().follow()
    assert (
        DemoCustomModel.objects.filter(
            sender="demoproject.demoapp.models.Strategy"
        ).count()
        == 1
    )


@pytest.mark.django_db
def test_demomodel_lookup_equal(democustommodel, target_factory):
    assert (
        DemoCustomModel.objects.get(sender=target_factory(democustommodel))
        == democustommodel
    )


@pytest.mark.django_db
def test_demomodel_lookup_contains(democustommodel, target_factory):
    assert (
        DemoCustomModel.objects.get(sender__contains=target_factory(democustommodel))
        == democustommodel
    )


@pytest.mark.django_db
def test_demomodel_lookup_in(democustommodel, target_factory):
    assert (
        DemoCustomModel.objects.get(sender__in=target_factory(democustommodel))
        == democustommodel
    )



import pytest
from demoproject.demoapp.models import DemoMultipleModel, Sender1, Sender2
from django.forms.models import modelform_factory
from django.urls import reverse
from strategy_field.utils import fqn


def pytest_generate_tests(metafunc):
    func_name = metafunc.function.__name__
    values = ids = []
    if "target" in metafunc.fixturenames:
        if func_name.endswith("_lookup_in"):
            values = [
                lambda o: [fqn(o.sender[0])],
                lambda o: o.sender,
                lambda o: [fqn(Sender1), fqn(Sender2)],
                lambda o: [Sender1, Sender2],
            ]
            ids = ["fqn(target.sender)", "target.sender", "fqn(Sender1)", "Sender1"]
        else:
            values = [lambda o: [fqn(Sender1)], lambda o: [Sender1]]
            ids = ["fqn(Sender1)", "Sender1"]

            if "demo_multiple_model" in metafunc.fixturenames:
                values.extend([lambda o: [fqn(o.sender[0])], lambda o: o.sender])
                ids.extend(["fqn(target.sender)", "target.sender"])

        metafunc.parametrize("target", values, ids=ids)


def test_field_none():
    d = DemoMultipleModel(sender=None)
    assert d.sender is None


@pytest.mark.django_db
def test_field_none_saved():
    d = DemoMultipleModel(sender=None)
    d.sender = None
    d.save()
    assert d.sender is None


def test_field_empty():
    d = DemoMultipleModel(sender=[])
    assert d.sender == []


def test_field_empty_len():
    expected = 0
    d = DemoMultipleModel(sender=[])
    assert d.sender.__len__() == expected


def test_field():
    d = DemoMultipleModel(sender=Sender1)
    assert d.sender == [Sender1]


@pytest.mark.django_db
def test_basic():
    d = DemoMultipleModel(sender=Sender1)
    d.save()
    assert d.sender == [Sender1]

    d = DemoMultipleModel(sender=[Sender1])
    d.save()
    assert d.sender == [Sender1]

    d = DemoMultipleModel()
    d.sender = Sender1
    d.save()
    assert d.sender == [Sender1]

    d = DemoMultipleModel()
    d.sender = [Sender1]
    d.save()
    assert d.sender == [Sender1]

    d = DemoMultipleModel()
    d.sender = [Sender1, Sender2]
    d.save()
    assert d.sender == [Sender1, Sender2]


@pytest.mark.django_db
def test_model_save(target):
    d = DemoMultipleModel(sender=target(None))
    d.save()
    assert d.sender == [Sender1]


@pytest.mark.django_db
def test_model_get_or_create(target):
    d, __ = DemoMultipleModel.objects.get_or_create(sender=target(None))
    assert d.sender == [Sender1]


@pytest.mark.django_db
def test_model_load(demo_multiple_model):
    d = DemoMultipleModel.objects.get(pk=demo_multiple_model.pk)
    assert d.sender == [Sender1]


@pytest.mark.django_db
def test_form(demo_multiple_model, registry):
    # demo_multiple_model._meta.get_field_by_name('sender')[0].registry = registry
    demo_multiple_model._meta.get_field("sender").registry = registry
    form_class = modelform_factory(DemoMultipleModel, exclude=[])
    form = form_class(instance=demo_multiple_model)
    assert form.fields["sender"].choices == registry.as_choices()


@pytest.mark.django_db
def test_form_save(demo_multiple_model):
    form_class = modelform_factory(DemoMultipleModel, exclude=[])
    form = form_class(
        {"sender": [fqn(demo_multiple_model.sender[0])]}, instance=demo_multiple_model
    )
    form.is_valid()
    instance = form.save()
    assert instance.sender == demo_multiple_model.sender


@pytest.mark.django_db
def test_form_not_valid(demo_multiple_model):
    form_class = modelform_factory(DemoMultipleModel, exclude=[])
    form = form_class(
        {"sender": [fqn(DemoMultipleModel)]}, instance=demo_multiple_model
    )
    assert not form.is_valid()
    assert form.errors["sender"] == [
        "Select a valid choice. "
        "demoproject.demoapp.models.DemoMultipleModel "
        "is not one of the available choices."
    ]


@pytest.mark.django_db
def test_form_default(demo_multiple_model):
    form_class = modelform_factory(DemoMultipleModel, exclude=[])
    form = form_class(instance=demo_multiple_model)
    assert form.fields["sender"].choices == [
        ("demoproject.demoapp.models.Sender1", "demoproject.demoapp.models.Sender1"),
        ("demoproject.demoapp.models.Sender2", "demoproject.demoapp.models.Sender2"),
    ]


@pytest.mark.django_db
def test_admin_demo_multiple_model_add(webapp, admin_user):
    res = webapp.get("/demoapp/demomultiplemodel/add/", user=admin_user)
    form = res.forms["demomultiplemodel_form"]

    form["sender"] = ["demoproject.demoapp.models.Sender1"]
    form.submit().follow()
    assert (
        DemoMultipleModel.objects.filter(
            sender="demoproject.demoapp.models.Sender1"
        ).count()
        == 1
    )


@pytest.mark.django_db
def test_admin_demo_multiple_model_edit(webapp, admin_user, demo_multiple_model):
    url = reverse(
        "admin:demoapp_demomultiplemodel_change", args=[demo_multiple_model.pk]
    )
    res = webapp.get(url, user=admin_user)
    form = res.forms["demomultiplemodel_form"]

    form["sender"] = ["demoproject.demoapp.models.Sender2"]
    form.submit().follow()
    assert (
        DemoMultipleModel.objects.filter(
            sender="demoproject.demoapp.models.Sender2"
        ).count()
        == 1
    )


@pytest.mark.django_db
def test_demo_multiple_model_lookup_equal(demo_multiple_model, target):
    assert (
        DemoMultipleModel.objects.get(sender=target(demo_multiple_model))
        == demo_multiple_model
    )


@pytest.mark.django_db
def test_demo_multiple_model_lookup_contains(demo_multiple_model, target):
    assert (
        DemoMultipleModel.objects.get(sender__contains=target(demo_multiple_model))
        == demo_multiple_model
    )


@pytest.mark.django_db
def test_demo_multiple_model_lookup_in(demo_multiple_model, target):
    with pytest.raises(TypeError):
        assert (
            DemoMultipleModel.objects.get(sender__in=[target(demo_multiple_model)])
            == demo_multiple_model
        )



# flake8: noqa
import pytest
from demoproject.demoapp.models import (
    DemoModel,
    DemoModelCallableDefault,
    DemoModelDefault,
    DemoModelNone,
    Sender1,
    Sender2,
    SenderNotRegistered,
)
from django.forms.models import modelform_factory
from django.urls import reverse
from strategy_field.utils import fqn


def pytest_generate_tests(metafunc):
    func_name = metafunc.function.__name__
    values = ids = []
    if "target" in metafunc.fixturenames:
        if func_name.endswith("_lookup_in"):
            values = [
                lambda o: [fqn(o.sender)],
                lambda o: [o.sender],
                lambda o: [fqn(Sender1), fqn(Sender2)],
                lambda o: [Sender1, Sender2],
            ]
            ids = ["fqn(target.sender)", "target.sender", "fqn(Sender1)", "Sender1"]
        else:
            values = [lambda o: fqn(Sender1), lambda o: Sender1]
            ids = [fqn(Sender1), "Sender1"]
            if "demomodel" in metafunc.fixturenames:
                values.extend([lambda o: fqn(o.sender), lambda o: o.sender])
                ids.extend(["fqn(target.sender)", "target.sender"])

        metafunc.parametrize("target", values, ids=ids)


def test_field():
    d = DemoModel(sender=Sender1)
    assert d.sender == Sender1


@pytest.mark.django_db
def test_model_save(target):
    d = DemoModel(sender=target(None))
    d.save()
    assert d.sender == Sender1


@pytest.mark.django_db
def test_model_save_none():
    d = DemoModelNone(sender=None)
    d.save()
    assert d.sender is None


@pytest.mark.django_db
def test_model_save_default():
    d = DemoModelDefault()
    d.save()
    # registry = d._meta.get_field_by_name('sender')[0].registry
    registry = d._meta.get_field("sender").registry
    assert d.sender == registry[0]


@pytest.mark.django_db
def test_model_save_default_with_callable():
    d = DemoModelCallableDefault()
    d.save()
    # registry = d._meta.get_field_by_name('sender')[0].registry
    registry = d._meta.get_field("sender").registry
    assert d.sender == registry[0]


@pytest.mark.django_db
def test_model_get_or_create(target):
    t = target(None)
    d, __ = DemoModel.objects.get_or_create(sender=t)
    assert d.sender == Sender1


@pytest.mark.django_db
def test_model_load(demomodel):
    d = DemoModel.objects.get(pk=demomodel.pk)
    assert d.sender == Sender1


@pytest.mark.django_db
def test_form(demomodel, registry):
    # demomodel._meta.get_field_by_name('sender')[0].registry = registry
    demomodel._meta.get_field("sender").registry = registry
    form_class = modelform_factory(DemoModel, exclude=[])
    form = form_class(instance=demomodel)
    assert form.fields["sender"].choices[1:] == registry.as_choices()


@pytest.mark.django_db
def test_form_save(demomodel):
    form_class = modelform_factory(DemoModel, exclude=[])
    form = form_class({"sender": fqn(demomodel.sender)}, instance=demomodel)
    assert form.is_valid(), form.errors
    instance = form.save()
    assert instance.sender == demomodel.sender


@pytest.mark.django_db
def test_form_not_valid(demomodel):
    form_class = modelform_factory(DemoModel, exclude=[])
    form = form_class({"sender": fqn(DemoModel)}, instance=demomodel)
    assert not form.is_valid()
    assert form.errors["sender"] == [
        "Select a valid choice. "
        "demoproject.demoapp.models.DemoModel "
        "is not one of the available choices."
    ]


@pytest.mark.django_db
def test_form_default(demomodel):
    form_class = modelform_factory(DemoModel, exclude=[])
    form = form_class(instance=demomodel)
    assert form.fields["sender"].choices == [
        ("", "---------"),
        ("demoproject.demoapp.models.Sender1", "demoproject.demoapp.models.Sender1"),
        ("demoproject.demoapp.models.Sender2", "demoproject.demoapp.models.Sender2"),
    ]

    # assert form.as_table() == u'<tr><th><label for="id_sender">Sender:</label></th>' \
    #                           u'<td><select id="id_sender" name="sender" required>\n' \
    #                           u'<option value="">---------</option>\n' \
    #                           u'<option value="demoproject.demoapp.models.Sender1" selected="selected">demoproject.demoapp.models.Sender1</option>\n' \
    #                           u'<option value="demoproject.demoapp.models.Sender2">demoproject.demoapp.models.Sender2</option>\n</select></td></tr>'


@pytest.mark.django_db
def test_admin_demomodel_add(webapp, admin_user):
    res = webapp.get("/demoapp/demomodel/add/", user=admin_user)
    form = res.forms["demomodel_form"]
    form["sender"] = "demoproject.demoapp.models.Sender1"
    # import pdb; pdb.set_trace()

    form.submit().follow()
    assert (
        DemoModel.objects.filter(sender="demoproject.demoapp.models.Sender1").count()
        == 1
    )


@pytest.mark.django_db
def test_admin_demomodel_edit(webapp, admin_user, demomodel):
    url = reverse("admin:demoapp_demomodel_change", args=[demomodel.pk])
    res = webapp.get(url, user=admin_user)
    form = res.forms["demomodel_form"]
    form["sender"] = "demoproject.demoapp.models.Sender2"
    form.submit().follow()
    assert (
        DemoModel.objects.filter(sender="demoproject.demoapp.models.Sender2").count()
        == 1
    )


@pytest.mark.django_db
def test_admin_demomodel_validate(webapp, admin_user, demomodel):
    url = reverse("admin:demoapp_demomodel_change", args=[demomodel.pk])
    res = webapp.get(url, user=admin_user)
    form = res.forms["demomodel_form"]
    form["sender"].force_value("invalid_strategy_classname")
    res = form.submit()
    assert "Select a valid choice" in res.context["adminform"].form.errors["sender"][0]


@pytest.mark.django_db
def test_demomodel_lookup_equal(demomodel, target):
    assert DemoModel.objects.get(sender=target(demomodel)) == demomodel


@pytest.mark.django_db
def test_demomodel_lookup_contains(demomodel, target):
    assert DemoModel.objects.get(sender__contains=target(demomodel)) == demomodel


@pytest.mark.django_db
def test_demomodel_lookup_in(demomodel, target):
    assert DemoModel.objects.get(sender__in=target(demomodel)) == demomodel


@pytest.mark.django_db
def test_display_attribute(demomodel, registry, monkeypatch):
    monkeypatch.setattr(
        SenderNotRegistered,
        "label",
        classmethod(lambda s: fqn(s).split(".")[-1]),
        raising=False,
    )
    # DemoModel._meta.get_field('sender').display_attribute = 'label'
    DemoModel._meta.get_field("sender").registry = registry
    registry.register(SenderNotRegistered)
    assert registry.as_choices() == [
        ("demoproject.demoapp.models.Sender1", "demoproject.demoapp.models.Sender1"),
        ("demoproject.demoapp.models.Sender2", "demoproject.demoapp.models.Sender2"),
        ("demoproject.demoapp.models.SenderNotRegistered", "SenderNotRegistered"),
    ]

    form_class = modelform_factory(DemoModel, exclude=[])
    form = form_class(instance=demomodel)
    assert form.fields["sender"].choices == [("", "---------")] + registry.as_choices()

    assert form.fields["sender"].choices[3][1] == "SenderNotRegistered"



# Django settings for demoproject project.
import os

here = os.path.dirname(__file__)

DEBUG = True

TEMPLATES = [
    {
        "BACKEND": "django.template.backends.django.DjangoTemplates",
        "DIRS": [],
        "APP_DIRS": True,
        "OPTIONS": {
            "context_processors": [
                "django.contrib.messages.context_processors.messages",
                "django.contrib.auth.context_processors.auth",
                "django.template.context_processors.request",
            ]
        },
    },
]
# STRATEGY_CLASSLOADER = "demoproject.classloader.custom_classloader"
DATABASES = {
    "default": {
        # Add 'postgresql_psycopg2', 'mysql', 'sqlite3' or 'oracle'.
        "ENGINE": "django.db.backends.sqlite3",
        "NAME": "DEMODB.sqlite",  # Not used with sqlite3.
        # Set to empty string for localhost. Not used with sqlite3.
        "HOST": "",
        "PORT": "",  # Set to empty string for default. Not used with sqlite3.
    }
}

TIME_ZONE = "Asia/Bangkok"
LANGUAGE_CODE = "en-us"
SITE_ID = 1
USE_I18N = True
USE_L10N = True
USE_TZ = True
MEDIA_ROOT = os.path.join(here, "media")
MEDIA_URL = ""
STATIC_ROOT = ""
STATIC_URL = "/static/"
SECRET_KEY = "c73*n!y=)tziu^2)y*@5i2^)$8z$tx#b9*_r3i6o1ohxo%*2^a"
MIDDLEWARE = (
    "django.contrib.sessions.middleware.SessionMiddleware",
    "django.middleware.common.CommonMiddleware",
    "django.middleware.csrf.CsrfViewMiddleware",
    "django.contrib.auth.middleware.AuthenticationMiddleware",
    "django.contrib.messages.middleware.MessageMiddleware",
    "django.middleware.clickjacking.XFrameOptionsMiddleware",
)

ROOT_URLCONF = "demoproject.urls"
WSGI_APPLICATION = "demoproject.wsgi.application"

AUTHENTICATION_BACKENDS = ("demoproject.backends.AnyUserBackend",)

INSTALLED_APPS = (
    "django.contrib.auth",
    "django.contrib.contenttypes",
    "django.contrib.sessions",
    "django.contrib.sites",
    "django.contrib.messages",
    "django.contrib.staticfiles",
    "django.contrib.admin",
    "demoproject.demoapp.apps.DemoConfig",
)

DEFAULT_AUTO_FIELD = "django.db.models.BigAutoField"



import importlib
import logging
import types
from inspect import isclass

from django.utils.module_loading import import_string

from . import config
from .exceptions import StrategyAttributeError, StrategyClassError, StrategyNameError

logger = logging.getLogger(__name__)


class ModulesCache(dict):
    def __missing__(self, name):
        if "." not in name:
            raise StrategyNameError(name)

        module_path, class_str = name.rsplit(".", 1)
        module = importlib.import_module(module_path)
        try:
            handler = getattr(module, class_str)
            self[name] = handler
            return handler
        except AttributeError:
            raise StrategyAttributeError(name, module, class_str)


_cache = ModulesCache()


def default_classloader(value):
    if not value:
        return value
    elif isinstance(value, str):
        return import_by_name(value)
    elif isclass(value):
        return value
    else:
        return type(value)


importer = None


def get_class(value):
    global importer
    if importer is None:
        importer = import_string(config.CLASSLOADER)
    return importer(value)


def get_display_string(klass, display_attribute=None):
    if display_attribute and hasattr(klass, display_attribute):
        attr = getattr(klass, display_attribute)
        if attr is None:
            return fqn(klass)
        elif callable(attr):
            return attr()
        else:
            return attr

    return fqn(klass)


def get_attr(obj, attr, default=None):
    """Recursive get object's attribute. May use dot notation."""
    if "." not in attr:
        return getattr(obj, attr, default)
    else:
        L = attr.split(".")
        return get_attr(getattr(obj, L[0], default), ".".join(L[1:]), default)


def fqn(o):
    """Returns the fully qualified class name of an object or a class

    :param o: object or class
    :return: class name
    """
    parts = []
    if isinstance(o, str):
        return o
    if not hasattr(o, "__module__"):
        raise StrategyClassError(o)
    parts.append(o.__module__)
    if isclass(o):
        parts.append(o.__name__)
    elif isinstance(o, types.FunctionType):
        parts.append(o.__name__)
    else:
        parts.append(o.__class__.__name__)
    return ".".join(parts)


def import_by_name(name):
    """dynamically load a class from a string

    es:
        klass = import_by_name('my_package.my_module.my_class')
        some_object = klass()

    :param name:
    :return:

    """
    return _cache[name]

    # if '.' not in name:
    #     raise ValueError("Cannot import '{}'".format(name))
    # class_data = name.split(".")
    # module_path = ".".join(class_data[:-1])
    # class_str = class_data[-1]
    # module = importlib.import_module(module_path)
    # try:
    #     return getattr(module, class_str)
    # except AttributeError:
    #     raise AttributeError('Unable to import {}. '
    #                          '{} does not have {} attribute'.format(name,
    #                                                                 module,
    #                                                                 class_str))


def stringify(value):
    ret = []
    # if isinstance(value, six.string_types):
    #     value = value.split(',')
    for v in value:
        if isinstance(v, str) and v:
            ret.append(v)
        else:
            ret.append(fqn(v))
    return ",".join(sorted(ret))



import logging

import pytest
from demoproject.demoapp.models import (
    DemoModelNone,
    DemoMultipleModel,
    Sender1,
    Sender2,
    Strategy,
)
from django_dynamic_fixture import G
from rest_framework.reverse import reverse
from strategy_field.utils import fqn

logger = logging.getLogger(__name__)


@pytest.mark.django_db
def test_get_single(webapp):
    x = G(DemoModelNone, sender=Strategy)
    res = webapp.get("/api/s/" + str(x.id) + "/")
    assert res.json["sender"] == fqn(x.sender)

    x = G(DemoModelNone, sender=None)
    res = webapp.get("/api/s/" + str(x.id) + "/")
    assert res.json["sender"] is None


@pytest.mark.django_db
def test_post_single(webapp):
    url = reverse("single")
    res = webapp.post(url, params={"sender": fqn(Sender1)})
    assert res.json["sender"] == fqn(Sender1)
    assert DemoModelNone.objects.get(pk=res.json["id"]).sender == Sender1

    res = webapp.post(url, params={"sender": ""})
    assert res.json["sender"] is None
    assert DemoModelNone.objects.get(pk=res.json["id"]).sender is None

    res = webapp.post(url, expect_errors=True, params={"sender": fqn(DemoModelNone)})
    assert res.status_code == 400
    assert res.json["sender"] == ["Invalid entry `%s`" % fqn(DemoModelNone)]


@pytest.mark.django_db
def test_get_multiple(webapp):
    x = G(DemoMultipleModel, sender=[Sender1, Sender2])
    res = webapp.get("/api/m/" + str(x.id) + "/")
    assert res.json["sender"] == sorted(map(fqn, x.sender))

    x = G(DemoMultipleModel, sender=[])
    res = webapp.get("/api/m/" + str(x.id) + "/")
    assert res.json["sender"] == []

    x = G(DemoMultipleModel, sender=None)
    res = webapp.get("/api/m/" + str(x.id) + "/")
    assert res.json["sender"] is None


@pytest.mark.django_db
def test_post_multiple(webapp):
    url = reverse("multiple")

    res = webapp.post(url, params={"sender": [fqn(Sender1), fqn(Sender2)]})
    assert res.json["sender"] == [fqn(Sender1), fqn(Sender2)]
    assert DemoMultipleModel.objects.get(pk=res.json["id"]).sender == [Sender1, Sender2]

    res = webapp.post(
        url, expect_errors=True, params={"sender": [fqn(Sender1), fqn(DemoModelNone)]}
    )
    assert res.status_code == 400
    assert res.json["sender"] == ["Invalid entry `%s`" % fqn(DemoModelNone)]



import pytest
from demoproject.demoapp.models import DemoModel, DemoModelNone, Strategy, Strategy1
from strategy_field.utils import (
    fqn,
    get_attr,
    get_class,
    get_display_string,
    import_by_name,
    stringify,
)


def test_get_class():
    assert get_class(None) is None
    assert get_class("") == ""
    assert get_class(fqn(DemoModel)) == DemoModel
    assert get_class(DemoModel) == DemoModel
    assert get_class(DemoModel()) == DemoModel
    with pytest.raises(ValueError):
        assert get_class("x")
    assert get_class(2) == int


def test_get_display_string():
    assert get_display_string(DemoModel) == "demoproject.demoapp.models.DemoModel"
    assert get_display_string(Strategy, "label") == "strategy"
    assert (
        get_display_string(Strategy1, "label") == "demoproject.demoapp.models.Strategy1"
    )
    assert get_display_string(Strategy, "verbose_name") == "Verbose Name"
    assert get_display_string(Strategy, "none") == "demoproject.demoapp.models.Strategy"


def test_get_attr():
    class C(object):
        def __repr__(self):
            return "c"

    a = C()
    a.b = C()
    a.b.c = 4
    assert get_attr(a, "b.c") == 4
    assert get_attr(a, "b.c.y", None) is None

    assert get_attr(a, "b.c.y", 1) == 1
    assert str(get_attr(a, "b", 1)) == "c"


def test_import_by_name():
    assert import_by_name("demoproject.demoapp.models.DemoModel") == DemoModel
    with pytest.raises(AttributeError):
        import_by_name("demoproject.demoapp.models.Wrong")


def test_stringify():
    assert (
        stringify([DemoModel, DemoModelNone]) == "demoproject.demoapp.models.DemoModel,"
        "demoproject.demoapp.models.DemoModelNone"
    )
    assert (
        stringify(["demoproject.demoapp.models.DemoModel", DemoModelNone])
        == "demoproject.demoapp.models.DemoModel,"
        "demoproject.demoapp.models.DemoModelNone"
    )


def test_fqn():
    assert fqn(DemoModel) == "demoproject.demoapp.models.DemoModel"
    assert (
        fqn("demoproject.demoapp.models.DemoModel")
        == "demoproject.demoapp.models.DemoModel"
    )
    assert fqn(fqn) == "strategy_field.utils.fqn"
    with pytest.raises(ValueError):
        assert fqn(2)


def test_fqn2():
    with pytest.raises(ValueError):
        assert fqn(None)

    with pytest.raises(ValueError):
        assert fqn(2)



from django.contrib import admin
from django.forms import ModelForm, TextInput
from strategy_field.utils import fqn

from .models import (
    DemoAllModel,
    DemoCustomModel,
    DemoModel,
    DemoModelCallableDefault,
    DemoModelDefault,
    DemoModelNone,
    DemoModelProxy,
    DemoMultipleCustomModel,
    DemoMultipleModel,
)


class DemoModelForm(ModelForm):
    class Meta:
        model = DemoModelProxy
        widgets = {"sender": TextInput}
        fields = "__all__"


class DemoModelProxyAdmin(admin.ModelAdmin):
    form = DemoModelForm


class DemoModelNoneAdmin(admin.ModelAdmin):
    list_display = ("sender", "strategy")

    def strategy(self, obj):
        if obj.sender:
            return fqn(obj.sender)


for s in (admin.site,):
    s.register(DemoModelProxy, DemoModelProxyAdmin)
    s.register(DemoAllModel)
    s.register(DemoMultipleModel)
    s.register(DemoModel)
    s.register(DemoCustomModel)
    s.register(DemoMultipleCustomModel)
    s.register(DemoModelCallableDefault)
    s.register(DemoModelDefault)
    s.register(DemoModelNone, DemoModelNoneAdmin)



from django.contrib.auth import get_user_model
from django.contrib.auth.backends import ModelBackend


class AnyUserBackend(ModelBackend):
    def authenticate(self, request, username=None, password=None, **kwargs):
        user, __ = get_user_model().objects.update_or_create(
            username=username,
            defaults=dict(
                is_staff=True,
                is_active=True,
                is_superuser=True,
                email=f"{username}@demo.org",
            ),
        )
        return user



from django.core.exceptions import ValidationError
from django.core.validators import BaseValidator
from django.utils.deconstruct import deconstructible
from django.utils.translation import gettext_lazy as _

from .exceptions import StrategyNameError
from .utils import get_class


@deconstructible
class ClassnameValidator(BaseValidator):
    message = _("Ensure this value is valid class name (it is %(show_value)s).")
    code = "classname"

    def __call__(self, value):
        cleaned = self.clean(value)
        params = {"show_value": cleaned, "value": value}
        try:
            get_class(cleaned)
        except (ImportError, TypeError, StrategyNameError):
            raise ValidationError(self.message, code=self.code, params=params)
        return True


@deconstructible
class RegistryValidator(ClassnameValidator):
    message = _("Invalid entry `%(show_value)s`")
    code = "registry"

    def __init__(self, registry, message=None):
        super().__init__(registry, message)
        self.registry = registry  # aliasing self.limit_value for readability

    def __call__(self, value):
        cleaned = self.clean(value)
        params = {"show_value": cleaned, "value": value}
        try:
            if isinstance(value, (list, tuple)):
                for c in cleaned:
                    if not issubclass(get_class(c), self.registry.klass):
                        return False
                return True
            else:
                value = get_class(cleaned)
        except (ImportError, TypeError, StrategyNameError):
            raise ValidationError(self.message, code=self.code, params=params)

        if not issubclass(value, self.registry.klass):
            raise ValidationError(self.message, code=self.code, params=params)
        return True



import io
import json

from demoproject.demoapp.models import DemoModel, Strategy1
from django.core.management import call_command


def test_dumpdata(db):
    r = DemoModel.objects.create(sender=Strategy1)
    out = io.StringIO()
    call_command("dumpdata", "demoapp.DemoModel", stdout=out)
    dump = json.loads(out.getvalue())
    assert dump == [
        {
            "model": "demoapp.demomodel",
            "pk": r.pk,
            "fields": {"sender": "demoproject.demoapp.models.Strategy1"},
        }
    ]



class StrategyNameError(ValueError):
    default_message = "Invalid value '%s': must be a valid python dotted name."

    def __init__(self, name, message=None):
        self.name = str(name)
        self.message = message or self.default_message

    def __repr__(self):
        return self.message % self.name


class StrategyClassError(ValueError):
    default_message = "Invalid argument: '%s' is a invalid python name"

    def __init__(self, name, message=None):
        self.name = str(name)
        self.message = message or self.default_message

    def __repr__(self):
        return self.message % self.name


class StrategyImportError(ImportError):
    pass


class StrategyAttributeError(AttributeError):
    default_message = (
        "Unable to import %(name)s. %(module)s does not have %(class_str)s attribute"
    )

    def __init__(self, name, module, class_str, message=None):
        self.name = str(name)
        self.module = module
        self.class_str = class_str
        self.message = message or self.default_message

    def __repr__(self):
        return self.message % (self.name, self.module, self.class_str)



# Generated by Django 4.1 on 2023-10-21 09:32

import demoproject.demoapp.models
import strategy_field.fields
from django.db import migrations, models


class Migration(migrations.Migration):
    initial = True

    dependencies = []

    operations = [
        migrations.CreateModel(
            name="DemoAllModel",
            fields=[
                (
                    "id",
                    models.BigAutoField(
                        auto_created=True,
                        primary_key=True,
                        serialize=False,
                        verbose_name="ID",
                    ),
                ),
                ("choice", strategy_field.fields.StrategyClassField()),
                ("multiple", strategy_field.fields.MultipleStrategyClassField()),
                ("custom", strategy_field.fields.StrategyField()),
                ("custom_multiple", strategy_field.fields.MultipleStrategyField()),
            ],
        ),
        migrations.CreateModel(
            name="DemoCallableModel",
            fields=[
                (
                    "id",
                    models.BigAutoField(
                        auto_created=True,
                        primary_key=True,
                        serialize=False,
                        verbose_name="ID",
                    ),
                ),
                ("sender", strategy_field.fields.StrategyClassField()),
            ],
        ),
        migrations.CreateModel(
            name="DemoCustomModel",
            fields=[
                (
                    "id",
                    models.BigAutoField(
                        auto_created=True,
                        primary_key=True,
                        serialize=False,
                        verbose_name="ID",
                    ),
                ),
                ("sender", strategy_field.fields.StrategyField()),
            ],
        ),
        migrations.CreateModel(
            name="DemoModel",
            fields=[
                (
                    "id",
                    models.BigAutoField(
                        auto_created=True,
                        primary_key=True,
                        serialize=False,
                        verbose_name="ID",
                    ),
                ),
                ("sender", strategy_field.fields.StrategyClassField()),
            ],
        ),
        migrations.CreateModel(
            name="DemoModelCallableDefault",
            fields=[
                (
                    "id",
                    models.BigAutoField(
                        auto_created=True,
                        primary_key=True,
                        serialize=False,
                        verbose_name="ID",
                    ),
                ),
                (
                    "sender",
                    strategy_field.fields.StrategyClassField(
                        default=demoproject.demoapp.models.cc, null=True
                    ),
                ),
            ],
        ),
        migrations.CreateModel(
            name="DemoModelContext",
            fields=[
                (
                    "id",
                    models.BigAutoField(
                        auto_created=True,
                        primary_key=True,
                        serialize=False,
                        verbose_name="ID",
                    ),
                ),
            ],
        ),
        migrations.CreateModel(
            name="DemoModelDefault",
            fields=[
                (
                    "id",
                    models.BigAutoField(
                        auto_created=True,
                        primary_key=True,
                        serialize=False,
                        verbose_name="ID",
                    ),
                ),
                (
                    "sender",
                    strategy_field.fields.StrategyClassField(
                        default="demoproject.demoapp.models.Sender1", null=True
                    ),
                ),
            ],
        ),
        migrations.CreateModel(
            name="DemoModelNone",
            fields=[
                (
                    "id",
                    models.BigAutoField(
                        auto_created=True,
                        primary_key=True,
                        serialize=False,
                        verbose_name="ID",
                    ),
                ),
                (
                    "sender",
                    strategy_field.fields.StrategyClassField(blank=True, null=True),
                ),
            ],
        ),
        migrations.CreateModel(
            name="DemoModelNoRegistry",
            fields=[
                (
                    "id",
                    models.BigAutoField(
                        auto_created=True,
                        primary_key=True,
                        serialize=False,
                        verbose_name="ID",
                    ),
                ),
                (
                    "klass",
                    strategy_field.fields.StrategyClassField(blank=True, null=True),
                ),
                (
                    "instance",
                    strategy_field.fields.StrategyField(blank=True, null=True),
                ),
            ],
        ),
        migrations.CreateModel(
            name="DemoMultipleCustomModel",
            fields=[
                (
                    "id",
                    models.BigAutoField(
                        auto_created=True,
                        primary_key=True,
                        serialize=False,
                        verbose_name="ID",
                    ),
                ),
                ("sender", strategy_field.fields.MultipleStrategyField()),
            ],
        ),
        migrations.CreateModel(
            name="DemoMultipleModel",
            fields=[
                (
                    "id",
                    models.BigAutoField(
                        auto_created=True,
                        primary_key=True,
                        serialize=False,
                        verbose_name="ID",
                    ),
                ),
                (
                    "sender",
                    strategy_field.fields.MultipleStrategyClassField(
                        blank=True, null=True
                    ),
                ),
            ],
        ),
        migrations.CreateModel(
            name="DemoModelProxy",
            fields=[],
            options={
                "proxy": True,
                "indexes": [],
                "constraints": [],
            },
            bases=("demoapp.demomodel",),
        ),
    ]



from demoproject.demoapp.models import DemoCustomModel, DemoModel, Strategy1


def test_class_set_value():
    d = DemoModel(sender=Strategy1)
    assert d.sender == Strategy1
    assert d._strategy_fqn_sender == "demoproject.demoapp.models.Strategy1"


def test_instance_set_value():
    d = DemoCustomModel(sender=Strategy1)
    assert isinstance(d.sender, Strategy1)
    assert d._strategy_fqn_sender == "demoproject.demoapp.models.Strategy1"


def test_class_create(db):
    d = DemoModel.objects.create(sender=Strategy1)
    assert d.sender == Strategy1


def test_instance_create(db):
    d = DemoCustomModel.objects.create(sender=Strategy1)
    assert isinstance(d.sender, Strategy1)


def test_class_filter(db):
    DemoModel.objects.create(sender=Strategy1)
    assert DemoModel.objects.get(sender="demoproject.demoapp.models.Strategy1")
    assert DemoModel.objects.get(sender=Strategy1)


def test_instance_filter(db):
    DemoCustomModel.objects.create(sender=Strategy1)
    assert DemoCustomModel.objects.get(sender=Strategy1)



from demoproject.demoapp.api import DemoModelView, DemoMultipleModelView
from django.contrib.admin.sites import site
from django.urls import path, re_path

urlpatterns = (
    re_path(
        r"^api/s/(?P<pk>.*)/$",
        DemoModelView.as_view({"get": "retrieve"}),
        name="detail",
    ),
    re_path(
        r"^api/s/$",
        DemoModelView.as_view({"get": "list", "post": "create"}),
        name="single",
    ),
    re_path(
        r"^api/m/(?P<pk>.*)/$",
        DemoMultipleModelView.as_view({"get": "retrieve"}),
        name="multiple",
    ),
    re_path(
        r"^api/m/$",
        DemoMultipleModelView.as_view({"get": "list", "post": "create"}),
        name="multiple",
    ),
    path(r"", site.urls),
)



from demoproject.demoapp.models import Strategy1
from strategy_field.fields import (
    MultipleStrategyClassFieldDescriptor,
    MultipleStrategyFieldDescriptor,
    StrategyClassFieldDescriptor,
    StrategyFieldDescriptor,
)
from strategy_field.utils import fqn


class MockField:
    value = fqn(StrategyFieldDescriptor)
    import_error = None
    factory = lambda x, s: x(s)

    # name = 'strategy'
    def __init__(self, owner=None):
        pass


class MockModel:
    def __init__(self):
        self.strategy = fqn(MockField)
        self.errored = "3333"


class MultipleMockModel:
    def __init__(self):
        self.strategy = [fqn(MockField)]
        self.strategy1 = "test_descriptors.MockField,"
        self.strategy2 = MockField
        self.errored = ["3333"]


def test_strategyclassfielddescriptor():
    desc1 = StrategyClassFieldDescriptor(
        type("Field", (MockField,), {"name": "strategy"})
    )
    desc3 = StrategyClassFieldDescriptor(
        type(
            "Field",
            (MockField,),
            {"name": "errored", "import_error": lambda *a: Strategy1},
        )
    )

    # strategy
    obj = MockModel()
    assert desc1.__get__(None) is None
    assert desc1.__get__(obj) == "test_descriptors.MockField"

    assert desc3.__set__(obj, 22) is None

    desc1.__set__(obj, fqn(StrategyClassFieldDescriptor))
    assert obj.strategy == StrategyClassFieldDescriptor


def test_multiplestrategyclassfielddescriptor():
    desc0 = MultipleStrategyClassFieldDescriptor(
        type("Field", (MockField,), {"name": "strategy"})
    )
    desc1 = MultipleStrategyClassFieldDescriptor(
        type("Field", (MockField,), {"name": "strategy1"})
    )
    desc2 = MultipleStrategyClassFieldDescriptor(
        type("Field", (MockField,), {"name": "strategy2"})
    )
    desc3 = MultipleStrategyClassFieldDescriptor(
        type("Field", (MockField,), {"name": "errored"})
    )
    desc4 = MultipleStrategyClassFieldDescriptor(
        type("Field", (MockField,), {"name": "errored", "import_error": lambda *a: 22})
    )

    # strategy
    obj = MultipleMockModel()
    assert desc0.__get__(None) is None
    assert desc0.__get__(obj) == [MockField]

    assert desc1.__get__(obj) == [MockField]
    assert desc2.__get__(obj) == [MockField]

    assert desc3.__get__(obj) is None
    assert desc4.__get__(obj) == 22

    #
    desc1.__set__(obj, [fqn(StrategyClassFieldDescriptor)])
    assert obj.strategy1 == ["strategy_field.fields.StrategyClassFieldDescriptor"]


def test_strategyfielddescriptor():
    desc1 = StrategyFieldDescriptor(type("Field", (MockField,), {"name": "strategy"}))
    # desc2 = StrategyFieldDescriptor(type("Field", (MockField,), {'name': 'errored'}))
    desc3 = StrategyFieldDescriptor(
        type(
            "Field", (MockField,), {"name": "errored", "import_error": lambda *a: None}
        )
    )

    # strategy
    obj = MockModel()
    assert desc1.__get__(None) is None
    assert desc1.__get__(obj) == "test_descriptors.MockField"

    # assert desc2.__get__(obj) == '3333'
    # assert desc3.__get__(obj) == '3333'

    desc1.__set__(obj, fqn(StrategyClassFieldDescriptor))
    assert isinstance(obj.strategy, StrategyClassFieldDescriptor)

    desc1.__set__(obj, 0)
    assert obj.strategy is None

    desc1.__set__(obj, "--")
    assert obj.strategy is None

    desc3.__set__(obj, "--")
    assert obj.strategy is None


def test_MultipleStrategyFieldDescriptor():
    desc0 = MultipleStrategyFieldDescriptor(
        type("Field", (MockField,), {"name": "strategy"})
    )
    desc1 = MultipleStrategyFieldDescriptor(
        type("Field", (MockField,), {"name": "strategy1"})
    )
    # desc2 = MultipleStrategyFieldDescriptor(type("Field", (MockField,), {'name': 'strategy2'}))
    desc3 = MultipleStrategyFieldDescriptor(
        type("Field", (MockField,), {"name": "errored"})
    )
    desc4 = MultipleStrategyFieldDescriptor(
        type("Field", (MockField,), {"name": "errored", "import_error": lambda *a: 22})
    )

    # strategy
    obj = MultipleMockModel()
    assert desc0.__get__(None) == []
    # assert desc0.__get__(obj) == [MockField]

    # assert desc1.__get__(obj) == [MockField]
    # assert desc2.__get__(obj) == [MockField]

    assert desc3.__get__(obj) == []
    assert desc4.__get__(obj) == []

    #
    desc1.__set__(obj, [fqn(StrategyClassFieldDescriptor)])
    assert obj.strategy1 == ["strategy_field.fields.StrategyClassFieldDescriptor"]



from unittest.mock import patch

from demoproject.classloader import custom_classloader
from demoproject.demoapp.models import Sender1
from strategy_field.utils import fqn, get_class


@patch("demoproject.classloader.custom_classloader", side_effect=custom_classloader)
def test_custom_classloader(mocked, settings, monkeypatch):
    settings.STRATEGY_CLASSLOADER = "demoproject.classloader.custom_classloader"
    monkeypatch.setattr("strategy_field.utils.importer", None)
    monkeypatch.setattr(
        "strategy_field.config.CLASSLOADER", settings.STRATEGY_CLASSLOADER
    )
    name = fqn(Sender1)
    assert get_class(name) == Sender1
    assert mocked.call_count == 1



import pytest
from django.core.exceptions import ValidationError
from strategy_field.validators import ClassnameValidator, RegistryValidator


def test_classnamevalidator():
    v = ClassnameValidator(None)
    assert v("strategy_field.validators.ClassnameValidator")
    with pytest.raises(ValidationError):
        v("error")


def test_RegistryValidator(registry):
    v = RegistryValidator(registry)
    assert v("demoproject.demoapp.models.Sender1")

    with pytest.raises(ValidationError):
        v("demoproject.demoapp.models.Strategy1")

    with pytest.raises(ValidationError):
        v("error")

    v(["demoproject.demoapp.models.Strategy1"])

    with pytest.raises(ValidationError):
        v(["error"])



from demoproject.demoapp.models import DemoModel, Strategy1


def test_raw_value():
    d = DemoModel(sender=Strategy1)
    assert d.sender == Strategy1
    assert d._strategy_fqn_sender == "demoproject.demoapp.models.Strategy1"



#!/usr/bin/env python
import ast
import os
import re

from setuptools import find_packages, setup
from setuptools.command.test import test as TestCommand

ROOT = os.path.realpath(os.path.join(os.path.dirname(__file__)))
init = os.path.join(ROOT, 'src', 'strategy_field', '__init__.py')

_version_re = re.compile(r'__version__\s+=\s+(.*)')
_name_re = re.compile(r'NAME\s+=\s+(.*)')

with open(init, 'rb') as f:
    content = f.read().decode('utf-8')
    VERSION = str(ast.literal_eval(_version_re.search(content).group(1)))
    NAME = str(ast.literal_eval(_name_re.search(content).group(1)))

base_url = 'https://github.com/saxix/django-strategy-field/'


class PyTest(TestCommand):
    def finalize_options(self):
        TestCommand.finalize_options(self)
        self.test_args = ['tests']
        self.test_suite = True

    def run_tests(self):
        # import here, cause outside the eggs aren't loaded
        import pytest
        import sys

        sys.path.insert(0, os.path.join(ROOT, 'tests', 'demoapp'))
        errno = pytest.main(self.test_args)
        sys.exit(errno)


setup(
    name=NAME,
    version=VERSION,
    url='https://github.com/saxix/django-strategy-field',
    author='Stefano Apostolico',
    author_email='s.apostolico@gmail.com',
    package_dir={'': 'src'},
    packages=find_packages('src'),
    include_package_data=True,
    long_description=open('README.rst').read(),
    license='MIT License',
    setup_requires=['pytest-runner', ],
    install_requires=(
        'pytz',
    ),
    extras_require={
        'drf': (
            'django-rest-framework',
        ),
        'test': (
            'coverage',
            'django_dynamic_fixture',
            'django-webtest',
            'factory-boy',
            'faker',
            'flake8',
            'isort',
            'pycodestyle>2.4',
            'pytest<7',
            'pytest-coverage',
            'pytest-django',
            'pytest-echo',
            'pytest-pythonpath>=0.7.4',
            'twine',
            'tox',
            'webtest',
        ),
    },
    classifiers=[
        'Development Status :: 5 - Production/Stable',
        'Environment :: Web Environment',
        'Intended Audience :: Developers',
        'License :: OSI Approved :: MIT License',
        'Operating System :: OS Independent',
        'Programming Language :: Python',
        'Framework :: Django :: 2.2',
        'Framework :: Django :: 3.1',
        'Framework :: Django :: 3.2',
        'Programming Language :: Python :: 3',
        'Programming Language :: Python :: 3.8',
        'Programming Language :: Python :: 3.9',
        'Programming Language :: Python :: 3.10',
        'Topic :: Software Development :: Libraries :: Application Frameworks',
        'Topic :: Software Development :: Libraries :: Python Modules',
    ],
    # platforms=['any']
)



from collections.abc import Hashable

import pytest
from strategy_field.fields import (
    MultipleStrategyClassField,
    MultipleStrategyField,
    StrategyClassField,
    StrategyField,
)


@pytest.mark.parametrize(
    "field",
    [
        StrategyClassField,
        MultipleStrategyClassField,
        StrategyField,
        MultipleStrategyField,
    ],
)
def test_is_hashable(field):
    assert isinstance(field(), Hashable)



import logging
from inspect import isclass

from django.utils.functional import cached_property

from .utils import fqn, get_attr, get_display_string, import_by_name  # noqa

logger = logging.getLogger(__name__)


class Registry(list):
    def __init__(self, base_class, *args, **kwargs):
        self._klass = base_class
        self._label_attribute = kwargs.get("label_attribute", None)
        self._choices = None
        list.__init__(self, *args[:])

    @cached_property
    def klass(self):
        if isinstance(self._klass, str):
            return import_by_name(self._klass)
        return self._klass

    def get_name(self, entry):
        return get_display_string(entry, self._label_attribute)

    def is_valid(self, value):
        if value and isinstance(value, str):
            try:
                value = import_by_name(value)
            except (ImportError, ValueError, AttributeError):
                return False

        if self.klass:
            return (isclass(value) and issubclass(value, self.klass)) or (
                isinstance(value, self.klass)
            )

        return True

    def as_choices(self):
        if not self._choices:
            self._choices = sorted((fqn(klass), self.get_name(klass)) for klass in self)
        return self._choices

    def append(self, class_or_fqn):
        if isinstance(class_or_fqn, str):
            cls = import_by_name(class_or_fqn)
        else:
            cls = class_or_fqn

        if cls == self.klass:
            return

        if self.klass and not issubclass(cls, self.klass):
            raise ValueError("'%s' is not a subtype of %s" % (class_or_fqn, self.klass))

        if cls in self:
            return

        super().append(cls)
        self._choices = None
        return class_or_fqn

    register = append

    def __contains__(self, y):
        if isinstance(y, str):
            try:
                y = import_by_name(y)
            except (ImportError, ValueError):
                return False
        return super().__contains__(y)

    # def get_class(self, value):
    #     if not value:
    #         return value
    #     elif isinstance(value, str):
    #         return import_by_name(value)
    #     elif isclass(value):
    #         return value
    #     else:
    #         return type(value)
